@@ -1,29 +1,28 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 from decimal import Decimal
-from trytond.model import Workflow, ModelView, fields
-from trytond.pyson import Eval
+from trytond.model import Workflow, ModelView, fields, Check
 from trytond.pool import Pool, PoolMeta
 from trytond.transaction import Transaction
 
 __all__ = ['Move']
-__metaclass__ = PoolMeta
 
 
-class Move:
+class Move(metaclass=PoolMeta):
     __name__ = 'stock.move'
-    fifo_quantity = fields.Float('FIFO Quantity',
-        digits=(16, Eval('unit_digits', 2)),
-        depends=['unit_digits'])
+    fifo_quantity = fields.Float('FIFO Quantity')
 
     @classmethod
     def __setup__(cls):
         super(Move, cls).__setup__()
+        cls._allow_modify_closed_period.add('fifo_quantity')
+
+        t = cls.__table__()
         cls._sql_constraints += [
-                ('check_fifo_quantity_out',
-                    'CHECK(quantity >= fifo_quantity)',
-                    'FIFO quantity can not be greater than quantity.'),
-                ]
+            ('check_fifo_quantity_out',
+                Check(t, t.quantity >= t.fifo_quantity),
+                'FIFO quantity can not be greater than quantity.'),
+            ]
         cls._error_messages.update({
                 'del_move_fifo': ('You can not delete move "%s" that is used '
                     'for FIFO cost price.'),
@@ -46,10 +45,11 @@ class Move:
         total_qty = Uom.compute_qty(self.uom, self.quantity,
             self.product.default_uom, round=False)
 
-        fifo_moves = self.product.template.get_fifo_move(total_qty)
+        fifo_moves = self.product.get_fifo_move(total_qty)
 
         cost_price = Decimal("0.0")
         consumed_qty = 0.0
+        to_save = []
         for move, move_qty in fifo_moves:
             consumed_qty += move_qty
 
@@ -61,55 +61,50 @@ class Move:
                     move.product.default_uom)
             cost_price += move_unit_price * Decimal(str(move_qty))
 
-            move.quantity = move_qty
-            move._update_product_cost_price('out')
-
             move_qty = Uom.compute_qty(self.product.default_uom, move_qty,
                     move.uom, round=False)
-            # Use write as move instance quantity was modified to call
-            # _update_product_cost_price
-            self.write([self.__class__(move.id)], {
-                    'fifo_quantity': (move.fifo_quantity or 0.0) + move_qty,
-                    })
+            move.fifo_quantity = (move.fifo_quantity or 0.0) + move_qty
+            to_save.append(move)
+        if to_save:
+            # TODO save in do method when product change
+            self.__class__.save(to_save)
 
         if Decimal(str(consumed_qty)) != Decimal("0"):
             cost_price = cost_price / Decimal(str(consumed_qty))
 
+        # Compute average cost price
+        unit_price = self.unit_price
+        self.unit_price = Uom.compute_price(
+            move.product.default_uom, cost_price, self.uom)
+        average_cost_price = self._compute_product_cost_price('out')
+        self.unit_price = unit_price
+
         if cost_price != Decimal("0"):
             digits = self.__class__.cost_price.digits
-            return cost_price.quantize(
+            cost_price = cost_price.quantize(
                 Decimal(str(10.0 ** -digits[1])))
         else:
-            return self.product.cost_price
+            cost_price = average_cost_price
+        return cost_price, average_cost_price
 
-    @classmethod
-    @ModelView.button
-    @Workflow.transition('done')
-    def do(cls, moves):
-        pool = Pool()
-        Date = pool.get('ir.date')
-
-        today = Date.today()
-        for move in moves:
-            if not move.effective_date:
-                move.effective_date = today
-            if (move.from_location.type in ('supplier', 'production')
-                    and move.to_location.type == 'storage'
-                    and move.product.cost_price_method == 'fifo'):
-                move._update_product_cost_price('in')
-            elif (move.to_location.type == 'supplier'
-                    and move.from_location.type == 'storage'
-                    and move.product.cost_price_method == 'fifo'):
-                move._update_product_cost_price('out')
-            elif (move.from_location.type == 'storage'
-                    and move.to_location.type != 'storage'
-                    and move.product.cost_price_method == 'fifo'):
-                cost_price = move._update_fifo_out_product_cost_price()
-                if move.cost_price is None:
-                    move.cost_price = cost_price
-            move.save()
-
-        super(Move, cls).do(moves)
+    def _do(self):
+        cost_price = super(Move, self)._do()
+        if (self.from_location.type in ('supplier', 'production')
+                and self.to_location.type == 'storage'
+                and self.product.cost_price_method == 'fifo'):
+            cost_price = self._compute_product_cost_price('in')
+        elif (self.to_location.type == 'supplier'
+                and self.from_location.type == 'storage'
+                and self.product.cost_price_method == 'fifo'):
+            cost_price = self._compute_product_cost_price('out')
+        elif (self.from_location.type == 'storage'
+                and self.to_location.type != 'storage'
+                and self.product.cost_price_method == 'fifo'):
+            fifo_cost_price, cost_price = (
+                self._update_fifo_out_product_cost_price())
+            if self.cost_price is None:
+                self.cost_price = fifo_cost_price
+        return cost_price
 
     @classmethod
     @ModelView.button
