@@ -1,14 +1,13 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import datetime
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 
-__all__ = ['Template']
-__metaclass__ = PoolMeta
+__all__ = ['Template', 'Product']
 
 
-class Template:
+class Template(metaclass=PoolMeta):
     __name__ = 'product.template'
 
     @classmethod
@@ -16,7 +15,11 @@ class Template:
         super(Template, cls).__setup__()
         new_sel = ('fifo', 'FIFO')
         if new_sel not in cls.cost_price_method.selection:
-            cls.cost_price_method.selection.append(new_sel)
+            cls.cost_price_method._field.selection.append(new_sel)
+
+
+class Product(metaclass=PoolMeta):
+    __name__ = 'product.product'
 
     def get_fifo_move(self, quantity=0.0):
         '''
@@ -37,15 +40,15 @@ class Template:
         location_ids = [l.id for l in locations]
         with Transaction().set_context(locations=location_ids,
                 stock_date_end=stock_date_end):
-            template = Template(self.id)
+            product = self.__class__(self.id)
         offset = 0
-        limit = Transaction().cursor.IN_MAX
-        avail_qty = template.quantity
+        limit = Transaction().database.IN_MAX
+        avail_qty = product.quantity
         fifo_moves = []
 
         while avail_qty > 0.0:
             moves = Move.search([
-                ('product.template.id', '=', template.id),
+                ('product', '=', product.id),
                 ('state', '=', 'done'),
                 ('from_location.type', '=', 'supplier'),
                 ], offset=offset, limit=limit,
@@ -57,7 +60,7 @@ class Template:
             for move in moves:
                 qty = Uom.compute_qty(move.uom,
                         move.quantity - move.fifo_quantity,
-                        template.default_uom, round=False)
+                        product.default_uom, round=False)
                 avail_qty -= qty
 
                 if avail_qty <= quantity:
@@ -71,3 +74,6 @@ class Template:
 
         fifo_moves.reverse()
         return fifo_moves
+
+    def recompute_cost_price_fifo(self):
+        return self.recompute_cost_price_average()
